from typing import List, Dict, Optional


class Entity:
    """
    Entity representing some unique object related to our data, e.g. a user, a product, a review, etc.

    Args:
        name: Name of the entity
        join_keys: List of feature names that uniquely identify this entity
        tags: Tags for the entity
    """

    def __init__(self, name: str, join_keys: List[str], tags: Optional[Dict[str, str]] = None):
        self._name = name
        self._join_keys = join_keys
        self._tags = tags or dict()

    def __repr__(self):
        return f'Entity(name={self.name}, join_keys={self.join_keys})'

    def __eq__(self, other):
        if not isinstance(other, Entity):
            return False

        if (self.name != other.name) or (self.join_keys != other.join_keys):
            return False

        return True

    def __hash__(self):
        return hash((self.name, self.join_keys))

    @property
    def name(self) -> str:
        return self._name

    @property
    def join_keys(self) -> List[str]:
        return self._join_keys

    @property
    def tags(self) -> Dict[str, str]:
        return self._tags



import logging
import os

import pandas as pd
import sqlalchemy.types
from sqlalchemy import create_engine

from snax.data_sources._oracle_utils import drop_table
from snax.data_sources.oracle_data_source import OracleDataSource
from snax.example_feature_repos.sports_feature_repo.nhl_games import data_path as original_nhl_data_path
from snax.example_feature_repos.users_with_nas_feature_repo.users_with_nas import \
    data_path as original_users_with_na_data_path

_ORACLE_CONNECTION_STRING = os.environ.get('ORACLE_CONNECTION_STRING')
_ORACLE_SCHEMA = os.environ.get('ORACLE_SCHEMA')
_ORACLE_NHL_TABLE = 'nhl_games'
_ORACLE_USERS_WITH_NAS_TABLE = 'users_with_nas'

_NHL_DATA = pd.read_csv(original_nhl_data_path)
_USERS_WITH_NA_DATA = pd.read_csv(original_users_with_na_data_path)

logger = logging.getLogger(__name__)


def none_if_oracle_not_set(func):
    def wrapper(*args, **kwargs):
        if _ORACLE_CONNECTION_STRING is None:
            return None
        return func(*args, **kwargs)

    return wrapper


@none_if_oracle_not_set
def create_nhl_games() -> OracleDataSource:
    engine = create_engine(_ORACLE_CONNECTION_STRING)
    _NHL_DATA.to_sql(
        con=engine, schema=_ORACLE_SCHEMA, name=_ORACLE_NHL_TABLE, if_exists='replace', index=False,
        dtype={'venue': sqlalchemy.types.VARCHAR(100)})  # Change to VARCHAR because default CLOB can't be index

    return OracleDataSource(
        name='nhl_games_in_memory',
        schema=_ORACLE_SCHEMA,
        table=_ORACLE_NHL_TABLE,
        engine=engine
    )


@none_if_oracle_not_set
def create_users_with_nas() -> OracleDataSource:
    engine = create_engine(_ORACLE_CONNECTION_STRING)
    _USERS_WITH_NA_DATA.to_sql(
        con=engine, schema=_ORACLE_SCHEMA, name=_ORACLE_USERS_WITH_NAS_TABLE, if_exists='replace', index=False,
        dtype={'string_id': sqlalchemy.types.VARCHAR(100)})  # Change to VARCHAR because default CLOB can't be index

    return OracleDataSource(
        name='users_with_nas_in_memory',
        schema=_ORACLE_SCHEMA,
        table=_ORACLE_USERS_WITH_NAS_TABLE,
        engine=engine
    )


@none_if_oracle_not_set
def create_users_with_nas_field_mapping() -> OracleDataSource:
    engine = create_engine(_ORACLE_CONNECTION_STRING)
    _USERS_WITH_NA_DATA.to_sql(
        con=engine, schema=_ORACLE_SCHEMA, name=_ORACLE_USERS_WITH_NAS_TABLE, if_exists='replace', index=False)

    return OracleDataSource(
        name='users_with_nas_in_memory',
        schema=_ORACLE_SCHEMA,
        table=_ORACLE_USERS_WITH_NAS_TABLE,
        engine=engine,
        field_mapping={
            'is_subscribed': 'issubscribed',
            'timestamp': 'time_stamp',
        }
    )


@none_if_oracle_not_set
def create_empty_data_source() -> OracleDataSource:
    engine = create_engine(_ORACLE_CONNECTION_STRING)
    drop_table(table=_ORACLE_NHL_TABLE, schema=_ORACLE_SCHEMA, engine=engine)

    return OracleDataSource(
        name='empty_oracle',
        schema=_ORACLE_SCHEMA,
        table='empty_table',
        engine=engine
    )



import os
from typing import Optional, Dict, List

import pandas as pd

from snax.data_sources.in_memory_data_source import InMemoryDataSource


class CsvDataSource(InMemoryDataSource):
    def __init__(self, name: str, csv_file_path: str, separator: str = ',',
                 field_mapping: Optional[Dict[str, str]] = None, tags: Optional[Dict] = None):
        super().__init__(name=name, data=None, field_mapping=field_mapping, tags=tags)
        self._csv_file_path = csv_file_path
        self._separator = separator

    @property
    def csv_file_path(self) -> str:
        return self._csv_file_path

    @property
    def separator(self) -> str:
        return self._separator

    def _load_data(self) -> pd.DataFrame:
        if os.path.exists(self._csv_file_path):
            data = pd.read_csv(self.csv_file_path, sep=self.separator)
            self._data = data
        else:
            self._data = pd.DataFrame()

    def _dump_data(self):
        self._data.to_csv(self.csv_file_path, sep=self.separator, index=False)

    def _select(self, columns: Optional[List[str]] = None, where_sql_query: Optional[str] = None) -> pd.DataFrame:
        if self._data is None:
            self._load_data()

        return super()._select(columns=columns, where_sql_query=where_sql_query)

    def _insert(self, key: List[str], columns: List[str], data: pd.DataFrame, if_exists: str = 'error'):
        if self._data is None:
            self._load_data()

        super()._insert(key=key, columns=columns, data=data, if_exists=if_exists)

        self._dump_data()



"""Source: https://docs.oracle.com/cd/A97630_01/appdev.920/a42525/apb.htm"""
ORACLE_RESERVED_WORDS = [
    'ACCESS', 'ELSE', 'MODIFY', 'START', 'ADD', 'EXCLUSIVE', 'NOAUDIT', 'SELECT', 'ALL', 'EXISTS', 'NOCOMPRESS',
    'SESSION', 'ALTER', 'FILE', 'NOT', 'SET', 'AND', 'FLOAT', 'NOTFOUND', 'SHARE', 'ANY', 'FOR', 'NOWAIT', 'SIZE',
    'ARRAYLEN', 'FROM', 'NULL', 'SMALLINT', 'AS', 'GRANT', 'NUMBER', 'SQLBUF', 'ASC', 'GROUP', 'OF', 'SUCCESSFUL',
    'AUDIT', 'HAVING', 'OFFLINE', 'SYNONYM', 'BETWEEN', 'IDENTIFIED', 'ON', 'SYSDATE', 'BY', 'IMMEDIATE', 'ONLINE',
    'TABLE', 'CHAR', 'IN', 'OPTION', 'THEN', 'CHECK', 'INCREMENT', 'OR', 'TO', 'CLUSTER', 'INDEX', 'ORDER', 'TRIGGER',
    'COLUMN', 'INITIAL', 'PCTFREE', 'UID', 'COMMENT', 'INSERT', 'PRIOR', 'UNION', 'COMPRESS', 'INTEGER', 'PRIVILEGES',
    'UNIQUE', 'CONNECT', 'INTERSECT', 'PUBLIC', 'UPDATE', 'CREATE', 'INTO', 'RAW', 'USER', 'CURRENT', 'IS', 'RENAME',
    'VALIDATE', 'DATE', 'LEVEL', 'RESOURCE', 'VALUES', 'DECIMAL', 'LIKE', 'REVOKE', 'VARCHAR', 'DEFAULT', 'LOCK', 'ROW',
    'VARCHAR2', 'DELETE', 'LONG', 'ROWID', 'VIEW', 'DESC', 'MAXEXTENTS', 'ROWLABEL', 'WHENEVER', 'DISTINCT', 'MINUS',
    'ROWNUM', 'WHERE', 'DROP', 'MODE', 'ROWS', 'WITH'
]

ORACLE_KEYWORDS = [
    'ADMIN', 'CURSOR', 'FOUND', 'MOUNT', 'AFTER', 'CYCLE', 'FUNCTION', 'NEXT', 'ALLOCATE', 'DATABASE', 'GO', 'NEW',
    'ANALYZE', 'DATAFILE', 'GOTO', 'NOARCHIVELOG', 'ARCHIVE', 'DBA', 'GROUPS', 'NOCACHE', 'ARCHIVELOG', 'DEC',
    'INCLUDING', 'NOCYCLE', 'AUTHORIZATION', 'DECLARE', 'INDICATOR', 'NOMAXVALUE', 'AVG', 'DISABLE', 'INITRANS',
    'NOMINVALUE', 'BACKUP', 'DISMOUNT', 'INSTANCE', 'NONE', 'BEGIN', 'DOUBLE', 'INT', 'NOORDER', 'BECOME', 'DUMP',
    'KEY', 'NORESETLOGS', 'BEFORE', 'EACH', 'LANGUAGE', 'NORMAL', 'BLOCK', 'ENABLE', 'LAYER', 'NOSORT', 'BODY', 'END',
    'LINK', 'NUMERIC', 'CACHE', 'ESCAPE', 'LISTS', 'OFF', 'CANCEL', 'EVENTS', 'LOGFILE', 'OLD', 'CASCADE', 'EXCEPT',
    'MANAGE', 'ONLY', 'CHANGE', 'EXCEPTIONS', 'MANUAL', 'OPEN', 'CHARACTER', 'EXEC', 'MAX', 'OPTIMAL', 'CHECKPOINT',
    'EXPLAIN', 'MAXDATAFILES', 'OWN', 'CLOSE', 'EXECUTE', 'MAXINSTANCES', 'PACKAGE', 'COBOL', 'EXTENT', 'MAXLOGFILES',
    'PARALLEL', 'COMMIT', 'EXTERNALLY', 'MAXLOGHISTORY', 'PCTINCREASE', 'COMPILE', 'FETCH', 'MAXLOGMEMBERS', 'PCTUSED',
    'CONSTRAINT', 'FLUSH', 'MAXTRANS', 'PLAN', 'CONSTRAINTS', 'FREELIST', 'MAXVALUE', 'PLI', 'CONTENTS', 'FREELISTS',
    'MIN', 'PRECISION', 'CONTINUE', 'FORCE', 'MINEXTENTS', 'PRIMARY', 'CONTROLFILE', 'FOREIGN', 'MINVALUE', 'PRIVATE',
    'COUNT', 'FORTRAN', 'MODULE', 'PROCEDURE', 'PROFILE', 'SAVEPOINT', 'SQLSTATE', 'TRACING', 'QUOTA', 'SCHEMA',
    'STATEMENT_ID', 'TRANSACTION', 'READ', 'SCN', 'STATISTICS', 'TRIGGERS', 'REAL', 'SECTION', 'STOP', 'TRUNCATE',
    'RECOVER', 'SEGMENT', 'STORAGE', 'UNDER', 'REFERENCES', 'SEQUENCE', 'SUM', 'UNLIMITED', 'REFERENCING', 'SHARED',
    'SWITCH', 'UNTIL', 'RESETLOGS', 'SNAPSHOT', 'SYSTEM', 'USE', 'RESTRICTED', 'SOME', 'TABLES', 'USING', 'REUSE',
    'SORT', 'TABLESPACE', 'WHEN', 'ROLE', 'SQL', 'TEMPORARY', 'WRITE', 'ROLES', 'SQLCODE', 'THREAD', 'WORK', 'ROLLBACK',
    'SQLERROR', 'TIME'
]

PL_SQL_RESERVED_WORDS = [
    'ABORT', 'BETWEEN', 'CRASH', 'DIGITS', 'ACCEPT', 'BINARY_INTEGER', 'CREATE', 'DISPOSE', 'ACCESS', 'BODY', 'CURRENT',
    'DISTINCT', 'ADD', 'BOOLEAN', 'CURRVAL', 'DO', 'ALL', 'BY', 'CURSOR', 'DROP', 'ALTER', 'CASE', 'DATABASE', 'ELSE',
    'AND', 'CHAR', 'DATA_BASE', 'ELSIF', 'ANY', 'CHAR_BASE', 'DATE', 'END', 'ARRAY', 'CHECK', 'DBA', 'ENTRY',
    'ARRAYLEN', 'CLOSE', 'DEBUGOFF', 'EXCEPTION', 'AS', 'CLUSTER', 'DEBUGON', 'EXCEPTION_INIT', 'ASC', 'CLUSTERS',
    'DECLARE', 'EXISTS', 'ASSERT', 'COLAUTH', 'DECIMAL', 'EXIT', 'ASSIGN', 'COLUMNS', 'DEFAULT', 'FALSE', 'AT',
    'COMMIT', 'DEFINITION', 'FETCH', 'AUTHORIZATION', 'COMPRESS', 'DELAY', 'FLOAT', 'AVG', 'CONNECT', 'DELETE', 'FOR',
    'BASE_TABLE', 'CONSTANT', 'DELTA', 'FORM', 'BEGIN', 'COUNT', 'DESC', 'FROM', 'FUNCTION', 'NEW', 'RELEASE', 'SUM',
    'GENERIC', 'NEXTVAL', 'REMR', 'TABAUTH', 'GOTO', 'NOCOMPRESS', 'RENAME', 'TABLE', 'GRANT', 'NOT', 'RESOURCE',
    'TABLES', 'GROUP', 'NULL', 'RETURN', 'TASK', 'HAVING', 'NUMBER', 'REVERSE', 'TERMINATE', 'IDENTIFIED',
    'NUMBER_BASE', 'REVOKE', 'THEN', 'IF', 'OF', 'ROLLBACK', 'TO', 'IN', 'ON', 'ROWID', 'TRUE', 'INDEX', 'OPEN',
    'ROWLABEL', 'TYPE', 'INDEXES', 'OPTION', 'ROWNUM', 'UNION', 'INDICATOR', 'OR', 'ROWTYPE', 'UNIQUE', 'INSERT',
    'ORDER', 'RUN', 'UPDATE', 'INTEGER', 'OTHERS', 'SAVEPOINT', 'USE', 'INTERSECT', 'OUT', 'SCHEMA', 'VALUES', 'INTO',
    'PACKAGE', 'SELECT', 'VARCHAR', 'IS', 'PARTITION', 'SEPARATE', 'VARCHAR2', 'LEVEL', 'PCTFREE', 'SET', 'VARIANCE',
    'LIKE', 'POSITIVE', 'SIZE', 'VIEW', 'LIMITED', 'PRAGMA', 'SMALLINT', 'VIEWS', 'LOOP', 'PRIOR', 'SPACE', 'WHEN',
    'MAX', 'PRIVATE', 'SQL', 'WHERE', 'MIN', 'PROCEDURE', 'SQLCODE', 'WHILE', 'MINUS', 'PUBLIC', 'SQLERRM', 'WITH',
    'MLSLABEL', 'RAISE', 'START', 'WORK', 'MOD', 'RANGE', 'STATEMENT', 'XOR', 'MODE', 'REAL', 'STDDEV', 'NATURAL',
    'RECORD', 'SUBTYPE'
]

ALL_SPECIAL_KEYWORDS = ORACLE_RESERVED_WORDS + ORACLE_KEYWORDS + PL_SQL_RESERVED_WORDS


def is_special(word: str):
    return word.upper() in ALL_SPECIAL_KEYWORDS



from pathlib import Path

from snax.data_sources.csv_data_source import CsvDataSource
from snax.entity import Entity
from snax.feature import Feature
from snax.feature_view import FeatureView
from snax.value_type import Int, String, Timestamp, Float

data_path = Path(__file__).parent.parent.parent / 'data' / 'game.csv'

# Define DataSource Objects --------------------------------------------------------------------------------------------
nhl_games_csv_data_source = CsvDataSource(
    name='nhl_games_csv',
    csv_file_path=data_path
)

# Define Entity Objects ------------------------------------------------------------------------------------------------
match_entity = Entity(
    name='game',
    join_keys=['game_id']
)

# Define FeatureView Objects -------------------------------------------------------------------------------------------
nhl_games_csv_feature_view = FeatureView(
    name='nhl_games_csv',
    entities=[match_entity],
    features=[
        Feature('game_id', Int),
        Feature('season', String),
        Feature('type', String),
        Feature('date_time_GMT', Timestamp),
        Feature('away_team_id', Int),
        Feature('home_team_id', Int),
        Feature('away_goals', Int),
        Feature('home_goals', Int),
        Feature('outcome', String),
        Feature('home_rink_side_start', String),
        Feature('venue', String),
        Feature('venue_link', String),
        Feature('venue_time_zone_id', String),
        Feature('venue_time_zone_offset', Float),
        Feature('venue_time_zone_tz', String),
    ],
    source=nhl_games_csv_data_source
)



import importlib
from pathlib import Path
from typing import List

from snax.data_sources.data_source_base import DataSourceBase
from snax.entity import Entity
from snax.feature_view import FeatureView
from snax.value_type import ValueType

DUMMY_ENTITY_ID = '__dummy_id'
DUMMY_ENTITY_NAME = '__dummy'
DUMMY_ENTITY = Entity(DUMMY_ENTITY_NAME, [DUMMY_ENTITY_ID], ValueType.STRING)


class RepoContents:
    def __init__(self, data_sources: List[DataSourceBase] = None, feature_views: List[FeatureView] = None,
                 entities: List[Entity] = None):
        self._data_sources = data_sources or []
        self._feature_views = feature_views or []
        self._entities = entities or []

    @property
    def data_sources(self) -> List[DataSourceBase]:
        return self._data_sources

    @property
    def feature_views(self) -> List[FeatureView]:
        return self._feature_views

    @property
    def entities(self) -> List[Entity]:
        return self._entities

    def get_data_source(self, name: str) -> DataSourceBase:
        for data_source in self.data_sources:
            if data_source.name == name:
                return data_source

    def get_feature_view(self, name: str) -> FeatureView:
        for feature_view in self.feature_views:
            if feature_view.name == name:
                return feature_view

    def get_entity(self, name: str) -> Entity:
        for entity in self.entities:
            if entity.name == name:
                return entity


def import_as_module(repo_file_path: str, repo_path: str):
    module_relative_path = str(repo_file_path).replace(str(repo_path), '').replace('/', '.').replace('.py', '')
    top_level_dir_name = repo_path.name
    module_path = '_feature_store_repo.' + top_level_dir_name + module_relative_path
    spec = importlib.util.spec_from_file_location(module_path, repo_file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def parse_repo(repo_path: str) -> RepoContents:
    # TODO: Implement also for git repos
    repo_path = Path(repo_path)
    repo_file_paths = {path.resolve() for path in repo_path.glob('**/*.py')
                       if path.is_file() and path.name != '__init__.py'}

    repo_contents = RepoContents()
    data_sources_set = set()
    for repo_file_paths in repo_file_paths:
        module = import_as_module(repo_file_paths, repo_path)

        for attr_name in dir(module):
            obj = getattr(module, attr_name)
            if isinstance(obj, DataSourceBase) and not any((obj is ds) for ds in repo_contents.data_sources):
                repo_contents.data_sources.append(obj)
                data_sources_set.add(obj)
            if isinstance(obj, FeatureView) and not any((obj is fv) for fv in repo_contents.feature_views):
                repo_contents.feature_views.append(obj)
            elif isinstance(obj, Entity) and not any((obj is entity) for entity in repo_contents.entities):
                repo_contents.entities.append(obj)

    repo_contents.entities.append(DUMMY_ENTITY)
    return repo_contents



from pathlib import Path

import pandas as pd
from pandas.testing import assert_frame_equal

from snax.example_feature_repos import sports_feature_repo
from snax.feature_store import FeatureStore


def test_initialize():
    sports_feature_repo_path = Path(sports_feature_repo.__file__).parent
    feature_store = FeatureStore(repo_path=sports_feature_repo_path)
    data_source_names = [data_source.name for data_source in feature_store.list_data_sources()]
    entity_names = [entity.name for entity in feature_store.list_entities()]
    feature_view_names = [feature_view.name for feature_view in feature_store.list_feature_views()]
    assert data_source_names == ['nhl_games_csv']
    assert entity_names == ['game', '__dummy']
    assert feature_view_names == ['nhl_games_csv']


def test_add_features_to_dataframe():
    sports_feature_repo_path = Path(sports_feature_repo.__file__).parent
    feature_store = FeatureStore(repo_path=sports_feature_repo_path)

    nhl_games_data_source = feature_store.get_data_source('nhl_games_csv')
    game_entity = feature_store.get_entity('game')
    where_sql_query = "(venue_time_zone_id=='America/Winnipeg' or venue_time_zone_id.str.contains('Vancouver')) and season==20172018"
    game_ids = nhl_games_data_source.select(
        columns=[game_entity],
        where_sql_query=where_sql_query
    )
    expected_game_ids = pd.DataFrame({'game_id': [2017020001, 2017020423]})

    assert game_ids.reset_index(drop=True).equals(expected_game_ids.reset_index(drop=True))

    feature_dataframe = feature_store.add_features_to_dataframe(
        dataframe=game_ids,
        feature_names=[
            'nhl_games_csv:outcome',
            'nhl_games_csv:venue'
        ],
        entity_name='game',
    )

    expected_feature_dataframe = pd.DataFrame({
        'game_id': {22: 2017020001, 40: 2017020423},
        'outcome': {22: 'away win REG', 40: 'home win REG'},
        'venue': {22: 'Bell MTS Place', 40: 'Rogers Arena'}
    })

    feature_dataframe.reset_index(inplace=True, drop=True)
    expected_feature_dataframe.reset_index(inplace=True, drop=True)

    assert_frame_equal(feature_dataframe, expected_feature_dataframe)



import logging
from typing import Optional, Dict, List

import pandas as pd
from pandas import MultiIndex
from sqlalchemy.engine import Engine

from snax.data_sources._oracle_utils import drop_table, add_columns, upsert, get_data_subset_in_db, \
    add_unique_constraint, ensure_table_exists, get_colnames, ensure_columns_exist, \
    pd_dataframe_to_comma_separated_tuples
from snax.data_sources.data_source_base import DataSourceBase

logger = logging.getLogger(__name__)


class OracleDataSource(DataSourceBase):
    """
    Oracle DB based data source

        Args:
            name: Name of the data source
            engine: SQLAlchemy engine to the Oracle DB
            schema: Name of the schema where the data is located
            table: Name of the table where the data is located
            field_mapping: A mapping from field names in this data source to feature names
            tags: Tags for the data source
    """

    def __init__(self, name: str, schema: str, table: str, engine: Engine,
                 field_mapping: Optional[Dict[str, str]] = None, tags: Optional[Dict] = None):
        super().__init__(name=name, field_mapping=field_mapping, tags=tags)
        self._engine = engine
        self._schema = schema
        self._table = table

        ensure_table_exists(self._table, self._schema, self._engine)

    def _select(self, columns: Optional[List[str]] = None, where_sql_query: Optional[str] = None) -> pd.DataFrame:
        joined_columns = ','.join(columns) if columns else '*'
        query = f'SELECT {joined_columns} FROM {self._schema}.{self._table}'
        if where_sql_query:
            query += f' WHERE {where_sql_query}'

        data = pd.read_sql(query, self._engine)
        return data

    def _insert(self, key: List[str], columns: List[str], data: pd.DataFrame, if_exists: str = 'error'):
        ensure_columns_exist(key, data.dtypes.to_dict(), self._table, self._schema, self._engine)
        add_unique_constraint(key, self._table, self._schema, self._engine)
        data = data.copy()

        existing_key_values = MultiIndex.from_frame(
            get_data_subset_in_db(data, key, self._table, self._schema, self._engine))
        inserted_key_values = MultiIndex.from_frame(data[key])

        inserted_in_existing = [item in existing_key_values for item in inserted_key_values]
        inserted_not_in_existing = [not item for item in inserted_in_existing]
        existing_columns = get_colnames(self._table, self._schema, self._engine)
        new_columns = [colname for colname in list(data.columns) if colname not in existing_columns]

        if if_exists == 'error':
            common_existing_and_inserted_columns = (set(data.columns) - set(key)).intersection(existing_columns)
            if any(inserted_in_existing) and len(common_existing_and_inserted_columns) > 0:
                raise ValueError(f'Data already exists in {self._schema}.{self._table}')

        if len(new_columns) > 0:
            add_columns(new_columns, data, self._table, self._schema, self._engine)
            upsert(key, new_columns, data, self._table, self._schema, self._engine)

        if any(inserted_not_in_existing):
            upsert(key, columns, data[inserted_not_in_existing], self._table, self._schema, self._engine)

        if if_exists == 'replace' and any(inserted_in_existing):
            upsert(key, columns, data[inserted_in_existing], self._table, self._schema, self._engine)

    def _where_sql_query_from_key_values(self, key: List[str], key_values: pd.DataFrame) -> str:
        query = f"({', '.join(key)}) IN ({pd_dataframe_to_comma_separated_tuples(key_values)})"
        return query

    def delete(self):
        drop_table(self._table, self._schema, self._engine)



import os

import pandas as pd
import pytest
from numpy import dtype
from sqlalchemy import create_engine, types, Table, Boolean, Float, Integer, String
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError

from snax.data_sources._oracle_utils import ensure_table_exists, drop_table, get_sqlalchemy_table, get_colnames, \
    get_base_column_types, add_unique_constraint, add_columns, get_data_subset_in_db, \
    sqlalchemy_column_type_to_base_type, \
    retype_dataframe, pd_series_to_comma_separated_tuple, escape_value, pd_dataframe_to_comma_separated_tuples, upsert
from snax.utils import frames_equal_up_to_row_ordering

ORACLE_CONNECTION_STRING = os.environ.get('ORACLE_CONNECTION_STRING')
ORACLE_SCHEMA = os.environ.get('ORACLE_SCHEMA')
SAMPLE_DATA_TABLE = 'sample_data'


def insert_test_data(engine: Engine) -> Engine:
    data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Mary'],
        'age': [30, 25, 40],
        'temperature': [98.6, 97.8, 98.3],
        'date_start': ['2020-01-01', '2020-01-02', '2020-01-03'],
        'time_start': ['12:00:00', '12:00:00', '12:00:00'],
        'is_active': [True, False, True],
        'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00'],
    })
    data.to_sql(
        con=engine, schema=ORACLE_SCHEMA, name=SAMPLE_DATA_TABLE, if_exists='replace', index=False,
        dtype={
            'first_name': types.VARCHAR(100),
            'age': types.INTEGER(),
            'temperature': types.FLOAT(),
            'date_start': types.VARCHAR(30),
            'time_start': types.VARCHAR(30),
            'datetime_start': types.VARCHAR(30)
        }
    )
    return engine


def get_engine() -> Engine:
    if ORACLE_CONNECTION_STRING is None:
        return None

    engine = create_engine(ORACLE_CONNECTION_STRING)
    drop_table(table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine)

    return engine


@pytest.fixture
def empty_engine() -> Engine:  # TODO: Clean-up after use
    engine = get_engine()
    if engine is None:
        return pytest.skip('ORACLE_CONNECTION_STRING is not set')
    return engine


@pytest.fixture
def engine():  # TODO: Clean-up after use
    engine = get_engine()
    if engine is None:
        return pytest.skip('ORACLE_CONNECTION_STRING is not set')

    insert_test_data(engine)
    return engine


def test_ensure_table_exists(empty_engine):
    ensure_table_exists(SAMPLE_DATA_TABLE, ORACLE_SCHEMA, empty_engine)
    result = empty_engine.execute(f'select * from {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}')
    assert result.fetchall() == []


def test_drop_table(engine):
    drop_table(SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    with pytest.raises(DatabaseError):
        engine.execute(f'select * from {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}')


def test_get_sqlalchemy_table(engine):
    table = get_sqlalchemy_table(SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    assert type(table) == Table
    assert table.name == SAMPLE_DATA_TABLE
    assert table.schema == ORACLE_SCHEMA
    assert table.columns[0].name == 'id'


def test_get_colnames(engine):
    colnames = get_colnames(SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    assert colnames == ['id', 'first_name', 'age', 'temperature', 'date_start', 'time_start', 'is_active',
                        'datetime_start']


def test_get_base_column_types(engine):
    column_types = get_base_column_types(SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    assert column_types == {
        'id': int,
        'first_name': str,
        'age': int,
        'temperature': float,
        'date_start': str,
        'time_start': str,
        'is_active': int,
        'datetime_start': str,
    }


def test_add_unique_constraint(engine):
    add_unique_constraint(['id', 'first_name'], SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    sql = f"SELECT * FROM user_cons_columns where CONSTRAINT_NAME = '{ORACLE_SCHEMA.upper()}_{SAMPLE_DATA_TABLE.upper()}_ID_FIRST_NAME_UNIQUE'"
    result = engine.execute(sql)
    expected_result = [
        (ORACLE_SCHEMA.upper(),
         f'{ORACLE_SCHEMA.upper()}_{SAMPLE_DATA_TABLE.upper()}_ID_FIRST_NAME_UNIQUE',
         'SAMPLE_DATA', 'ID', 1),
        (ORACLE_SCHEMA.upper(),
         f'{ORACLE_SCHEMA.upper()}_{SAMPLE_DATA_TABLE.upper()}_ID_FIRST_NAME_UNIQUE',
         'SAMPLE_DATA',
         'FIRST_NAME', 2)]
    assert set(result) == set(expected_result)


def test_add_columns(engine):
    data = pd.DataFrame({
        'middle_name': ['Carlos', 'Ariane', 'Debbie'],
        'height': [1.75, 1.60, 1.80],
        'weight': [70, 65, 80],
    })
    add_columns(
        columns=['middle_name', 'height', 'weight'], data=data,
        table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )
    data = pd.read_sql(f'select * from {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', engine)

    assert 'middle_name' in data.columns
    assert 'height' in data.columns
    assert 'weight' in data.columns


def test_upsert_with_adding_new_rows(engine):
    new_data = pd.DataFrame({
        'id': [4, 5, 6],
        'first_name': ['John', 'Jane', 'Mary'],
        'age': [30, 25, 40],
        'temperature': [98.6, 97.8, 98.3],
        'date_start': ['2020-01-01', '2020-01-02', '2020-01-03'],
        'time_start': ['12:00:00', '12:00:00', '12:00:00'],
        'is_active': [True, False, True],
        'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00'],
    })

    upsert(
        key=['id', 'first_name'],
        columns=['age', 'temperature', 'date_start', 'time_start', 'is_active', 'datetime_start'],
        data=new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )

    altered_data = pd.read_sql(sql=f'SELECT * FROM {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', con=engine)

    expected_altered_data = pd.DataFrame(
        {'id': [1, 2, 3, 6, 4, 5],
         'first_name': ['John', 'Jane', 'Mary', 'Mary', 'John', 'Jane'],
         'age': [30, 25, 40, 40, 30, 25],
         'temperature': [98.6, 97.8, 98.3, 98.3, 98.6, 97.8],
         'date_start': ['2020-01-01', '2020-01-02', '2020-01-03', '2020-01-03', '2020-01-01', '2020-01-02'],
         'time_start': ['12:00:00', '12:00:00', '12:00:00', '12:00:00', '12:00:00', '12:00:00'],
         'is_active': [1, 0, 1, 1, 1, 0],
         'datetime_start': [
             '2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00', '2020-01-03 12:00:00',
             '2020-01-01 12:00:00', '2020-01-02 12:00:00']})

    assert frames_equal_up_to_row_ordering(altered_data, expected_altered_data)


def test_upsert_with_adding_new_rows_with_missings(engine):
    new_data = pd.DataFrame({
        'id': [4, 5, 6],
        'first_name': ['John', 'Jane', 'Mary'],
        'age': [30, 25, 40],
    })

    upsert(
        key=['id', 'first_name'],
        columns=['age'],
        data=new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )

    altered_data = pd.read_sql(sql=f'SELECT * FROM {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', con=engine)

    expected_altered_data = pd.DataFrame(
        {'id': [1, 2, 3, 6, 4, 5],
         'first_name': ['John', 'Jane', 'Mary', 'Mary', 'John', 'Jane'],
         'age': [30, 25, 40, 40, 30, 25],
         'temperature': [98.6, 97.8, 98.3, None, None, None],
         'date_start': ['2020-01-01', '2020-01-02', '2020-01-03', None, None, None],
         'time_start': ['12:00:00', '12:00:00', '12:00:00', None, None, None],
         'is_active': [1.0, 0.0, 1.0, None, None, None],
         'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00', None, None, None]}
    )

    assert frames_equal_up_to_row_ordering(altered_data, expected_altered_data)


def test_upsert_replacing_old_rows(engine):
    new_data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Mary'],
        'is_active': [True, True, True],
        'temperature': [98.6, 97.8, 98.3],
    })

    upsert(
        key=['id', 'first_name'],
        columns=['temperature', 'is_active'],
        data=new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )

    altered_data = pd.read_sql(sql=f'SELECT * FROM {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', con=engine)
    expected_altered_data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Mary'],
        'age': [30, 25, 40],
        'temperature': [98.6, 97.8, 98.3],
        'date_start': ['2020-01-01', '2020-01-02', '2020-01-03'],
        'time_start': ['12:00:00', '12:00:00', '12:00:00'],
        'is_active': [1, 1, 1],
        'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00']
    })

    assert frames_equal_up_to_row_ordering(altered_data, expected_altered_data)


def test_upsert_inserting_new_columns(engine):
    new_data = pd.DataFrame({
        'id': [1, 2],
        'first_name': ['John', 'Jane'],
        'foobar': [1, 2],
        'barfoo': [3, 4]
    })

    add_columns(['foobar', 'barfoo'], new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine)
    upsert(
        key=['id', 'first_name'],
        columns=['foobar', 'barfoo'],
        data=new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )

    altered_data = pd.read_sql(sql=f'SELECT * FROM {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', con=engine)
    expected_altered_data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Mary'],
        'age': [30, 25, 40],
        'temperature': [98.6, 97.8, 98.3],
        'date_start': ['2020-01-01', '2020-01-02', '2020-01-03'],
        'time_start': ['12:00:00', '12:00:00', '12:00:00'], 'is_active': [1, 0, 1],
        'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00'],
        'foobar': ['1', '2', None],
        'barfoo': ['3', '4', None]
    })

    assert frames_equal_up_to_row_ordering(altered_data, expected_altered_data)


def test_upsert_single_id_column(engine):
    new_data = pd.DataFrame({
        'id': [1, 2, 5],
        'first_name': ['John', 'Jane', 'Mary']
    })

    upsert(
        key=['id'],
        columns=['first_name'],
        data=new_data, table=SAMPLE_DATA_TABLE, schema=ORACLE_SCHEMA, engine=engine
    )

    altered_data = pd.read_sql(sql=f'SELECT * FROM {ORACLE_SCHEMA}.{SAMPLE_DATA_TABLE}', con=engine)
    expected_altered_data = pd.DataFrame({
        'id': [1, 2, 3, 5],
        'first_name': ['John', 'Jane', 'Mary', 'Mary'],
        'age': [30.0, 25.0, 40.0, None],
        'temperature': [98.6, 97.8, 98.3, None],
        'date_start': ['2020-01-01', '2020-01-02', '2020-01-03', None],
        'time_start': ['12:00:00', '12:00:00', '12:00:00', None],
        'is_active': [1.0, 0.0, 1.0, None],
        'datetime_start': ['2020-01-01 12:00:00', '2020-01-02 12:00:00', '2020-01-03 12:00:00', None]
    })
    assert frames_equal_up_to_row_ordering(altered_data, expected_altered_data)


def test_get_data_subset_in_db(engine):
    data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Jack'],
    })
    data_subset = get_data_subset_in_db(data, ['id', 'first_name'], SAMPLE_DATA_TABLE, ORACLE_SCHEMA, engine)
    expected_data_subset = pd.DataFrame({
        'id': [1, 2],
        'first_name': ['John', 'Jane']
    })
    assert data_subset.equals(expected_data_subset)


def test_sqlalchemy_column_type_to_base_type():
    assert sqlalchemy_column_type_to_base_type(String()) == str
    assert sqlalchemy_column_type_to_base_type(Integer()) == int
    assert sqlalchemy_column_type_to_base_type(Float()) == float
    assert sqlalchemy_column_type_to_base_type(Boolean()) == int


def test_retype_dataframe():
    colname_to_type = {'a': str, 'b': int, 'c': float}
    data = pd.DataFrame({
        'a': ['a', 'b', 'c'],
        'b': ['1', '2', '3'],
        'c': ['1.0', '2.0', '3.0']
    })
    retyped_data = retype_dataframe(colname_to_type, data)
    assert retyped_data.dtypes.to_dict() == {'a': dtype('O'), 'b': dtype('int64'), 'c': dtype('float64')}


def test_escape_value():
    assert escape_value(True) == '1'
    assert escape_value(False) == '0'
    assert escape_value(1) == '1'
    assert escape_value(0) == '0'
    assert escape_value(1.0) == '1.0'
    assert escape_value(0.0) == '0.0'
    assert escape_value('abc') == "'abc'"
    assert escape_value('') == "''"
    assert escape_value(None) == 'null'


def test_pd_series_to_comma_separated_tuple():
    series = pd.Series({'a': 1, 'b': 5.6, 'c': True, 'd': 'str'})
    assert pd_series_to_comma_separated_tuple(series) == "(1, 5.6, 1, 'str')"


def test_pd_dataframe_to_comma_separated_tuples():
    data = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [5.6, 7.8, 9.0],
        'c': [True, False, True],
        'd': ['str', 'str2', 'str3']
    })
    assert pd_dataframe_to_comma_separated_tuples(data) == \
           "(1, 5.6, 1, 'str'), (2, 7.8, 0, 'str2'), (3, 9.0, 1, 'str3')"



import os.path
from pathlib import Path

import pandas as pd

from snax.utils import copy_to_temp, frames_equal_up_to_row_ordering

dummy_directory_path = Path(__file__).parent / 'test_data' / 'dummy_directory'
dummy_file_path = dummy_directory_path / 'dummy_file.txt'
empty_dummy_directory_path = dummy_directory_path / 'empty_dummy_directory'


def test_copy_file():
    temp_path = copy_to_temp(dummy_file_path)
    assert os.path.exists(temp_path)


def test_copy_empty_dir():
    gitkeep_file_path = empty_dummy_directory_path / '.gitkeep'
    if os.path.exists(gitkeep_file_path):
        os.remove(gitkeep_file_path)

    temp_path = copy_to_temp(empty_dummy_directory_path)
    assert os.path.exists(temp_path)


def test_copy_non_empty_dir():
    temp_path = copy_to_temp(dummy_directory_path)
    assert os.path.exists(temp_path)
    assert os.path.isdir(temp_path)
    assert sorted(os.listdir(temp_path)) == ['dummy_file.txt', 'empty_dummy_directory']


def test_frames_equal_up_to_row_ordering_frames_equal_shuffled():
    df1 = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    df2 = pd.DataFrame({'a': [2, 1, 3], 'b': [5, 4, 6]})
    assert frames_equal_up_to_row_ordering(df1, df2)


def test_frames_equal_up_to_row_ordering_frames_equal():
    df1 = pd.DataFrame({'a': [1, 2, 3], 'b': ["a", "b", "c"], "c": [True, False, True], "d": [1.0, 2.0, 3.0]})
    df2 = pd.DataFrame({'a': [1, 2, 3], 'b': ["a", "b", "c"], "c": [True, False, True], "d": [1.0, 2.0, 3.0]})
    assert frames_equal_up_to_row_ordering(df1, df2)


def test_frames_equal_up_to_row_ordering_frames_not_equal():
    df1 = pd.DataFrame({'a': [1, 2, 3], 'b': ["a", "b", "c"], "c": [True, False, True], "d": [1.0, 2.0, 3.0]})
    df2 = pd.DataFrame({'a': [1, 2, 3], 'b': ["a", "b", "c"], "c": [False, False, True], "d": [1.0, 2.0, 3.0]})
    assert not frames_equal_up_to_row_ordering(df1, df2)



from datetime import datetime

import pandas as pd
import pytest
from sqlalchemy import Boolean, Integer

from snax.data_sources.examples.in_memory import create_users_with_nas
from snax.entity import Entity
from snax.feature import Feature
from snax.feature_view import FeatureView
from snax.value_type import String, Timestamp, Float


@pytest.fixture
def feature_view_users_with_nas():
    return FeatureView(
        name='users_with_nas',
        entities=[Entity('user', join_keys=['id']), Entity('full_user', join_keys=['id', 'string_id'])],
        features=[
            Feature(name='first_name', dtype=String),
            Feature(name='timestamp', dtype=Timestamp),
            Feature(name='is_subscribed', dtype=Boolean),
            Feature(name='age', dtype=Float),
            Feature(name='children', dtype=Integer)
        ],
        source=create_users_with_nas(),
    )


def test_add_feature_to_dataframe_on_single_key_entity(feature_view_users_with_nas: FeatureView):
    entity_dataframe = pd.DataFrame({
        'id': [1, 2, 3],
    })
    feature_values = feature_view_users_with_nas.add_features_to_dataframe(
        dataframe=entity_dataframe,
        feature_names=['first_name', 'timestamp', 'is_subscribed', 'age', 'children'],
        entity_name='user',
    )
    expected_feature_values = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['Cirillo', 'Codi', 'Marion'],
        'timestamp': [
            datetime(2021, 9, 4, 10, 41, 25),
            datetime(2021, 12, 24, 15, 17, 57),
            datetime(2021, 12, 9, 2, 56, 18)
        ],
        'is_subscribed': [True, True, False],
        'age': [45.8, 4.9, None],
        'children': [1.0, 3.0, 4.0]}
    )

    feature_values.reset_index(inplace=True, drop=True)
    expected_feature_values.reset_index(inplace=True, drop=True)

    assert feature_values.equals(expected_feature_values)


def test_add_feature_to_dataframe_on_multi_key_entity(feature_view_users_with_nas: FeatureView):
    entity_dataframe = pd.DataFrame({
        'id': [2, 3, 5],
        'string_id': ['b', 'c', 'ef'],
    })
    feature_values = feature_view_users_with_nas.add_features_to_dataframe(
        dataframe=entity_dataframe,
        feature_names=['first_name', 'timestamp', 'is_subscribed', 'age', 'children'],
        entity_name='full_user',
    )

    expected_feature_values = pd.DataFrame({
        'id': [2, 3, 5],
        'string_id': ['b', 'c', 'ef'],
        'first_name': ['Codi', 'Marion', None],
        'timestamp': [
            datetime(2021, 12, 24, 15, 17, 57),
            datetime(2021, 12, 9, 2, 56, 18),
            datetime(2022, 3, 30, 17, 29, 16)
        ],
        'is_subscribed': [True, False, False],
        'age': [4.9, None, 62.0],
        'children': [3.0, 4.0, 2.0]
    })

    feature_values.reset_index(inplace=True, drop=True)
    expected_feature_values.reset_index(inplace=True, drop=True)

    assert feature_values.equals(expected_feature_values)



import enum


class ValueType(enum.Enum):
    UNKNOWN = 0
    STRING = 1
    INT = 2
    FLOAT = 3
    BOOL = 4
    TIMESTAMP = 5
    STRING_LIST = 6
    INT_LIST = 7
    FLOAT_LIST = 8
    BOOL_LIST = 9
    TIMESTAMP_LIST = 10
    NULL = 11

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if not isinstance(other, ValueType):
            return False
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    @staticmethod
    def from_string(value):
        try:
            return ValueType[value]
        except KeyError:
            return ValueType.UNKNOWN


Unknown = ValueType.UNKNOWN
String = ValueType.STRING
Int = ValueType.INT
Float = ValueType.FLOAT
Bool = ValueType.BOOL
Timestamp = ValueType.TIMESTAMP
StringList = ValueType.STRING_LIST
IntList = ValueType.INT_LIST
FloatList = ValueType.FLOAT_LIST
BoolList = ValueType.BOOL_LIST
TimestampList = ValueType.TIMESTAMP_LIST
Null = ValueType.NULL



from abc import ABC
from typing import Optional, Dict, List

import pandas as pd

from snax.column_like import ColumnLike, get_features_names

_VALID_IF_EXISTS_OPTIONS = ['error', 'ignore', 'replace']


class DataSourceBase(ABC):
    """
    Representation of a single table of data stored in some database
    It's main responsibility is to provide a data-frame (not necessarily with the correct column types) that represents
    (subset) of the underlying data
    It's secondary responsibility can be to provide a way how to modify the underlying data given a data-frame

    Args:
        name: Name of the data source
        field_mapping: A mapping from field names in this data source to feature names
        tags: Tags for the data source
    """

    def __init__(self, name: str, field_mapping: Optional[Dict[str, str]] = None, tags: Optional[Dict] = None):
        self._name = name
        self._field_mapping = field_mapping or dict()
        self._tags = tags or dict()

    def __repr__(self):
        return f'DataSource(name={self.name})'

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if (self.name != other.name):
            return False

        return True

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def field_mapping(self) -> Dict[str, str]:
        return self._field_mapping

    @property
    def tags(self) -> Dict:
        return self._tags

    def select(self, columns: Optional[List[ColumnLike]] = None, key: Optional[List[ColumnLike]] = None,
               key_values: Optional[pd.DataFrame] = None, where_sql_query: Optional[str] = None) -> pd.DataFrame:
        """
        Select a subset of the underlying data
        This can be done either by specifying the key and key_values or by specifying the where_sql_query

        Args:
            columns: Entities, Features or feature names to select, if None, all columns are selected
            key: List of column names giving unique constraint on a row in the data source
            key_values: Data frame with the key values
            where_sql_query: Optional filter query to apply to the selection, for now language depends on the data source

        Returns:
            A DataFrame containing the selected data
        """
        if (key is not None or key_values is not None) and where_sql_query is not None:
            raise ValueError('Cannot specify both key, key_values and where_sql_query')
        if (key is None and key_values is not None) or (key is not None and key_values is None):
            raise ValueError('Must specify both key and key_values')

        if key is not None and key_values is not None:
            string_key = self._column_likes_to_colnames(key)
            where_sql_query = self._where_sql_query_from_key_values(string_key, key_values)

        string_columns = self._column_likes_to_colnames(columns)
        selected_data = self._select(string_columns, where_sql_query)
        selected_data.rename(columns=self._field_mapping, inplace=True)
        return selected_data

    def insert(self, key: List[ColumnLike], columns: List[ColumnLike], data: pd.DataFrame, if_exists: str = 'error'):
        """
        Insert feature values corresponding to the given keys into the data source

        Args:
            key: List of column names giving unique constraint on a row in the data source
            columns: List of column names to use in the insert
            data: Data to insert, the column names must correspond to feature names (not to the
                data sources' underlying field names)
            if_exists: What to do if the data source already contains non-empty data for the given keys
                can be one of 'error', 'replace', 'ignore'

        Returns:
            None
        """
        if if_exists not in _VALID_IF_EXISTS_OPTIONS:
            raise ValueError(f'if_exists must be one of {_VALID_IF_EXISTS_OPTIONS}')

        string_key = self._column_likes_to_colnames(key)
        string_columns = self._column_likes_to_colnames(columns)
        self._insert(string_key, string_columns, data.rename(columns=self._inverse_field_mapping), if_exists)

    def _where_sql_query_from_key_values(self, key: List[str], key_values: pd.DataFrame) -> str:
        raise NotImplementedError('Has to be overridden by subclass')

    def _select(self, columns: Optional[List[str]] = None, where_sql_query: Optional[str] = None) -> pd.DataFrame:
        raise NotImplementedError('Has to be overridden by subclass')

    def _insert(self, key: List[str], columns: List[str], data: pd.DataFrame, if_exists: str = 'error'):
        raise NotImplementedError('Has to be overridden by subclass')

    @property
    def _inverse_field_mapping(self) -> Dict:
        if not hasattr(self, '_inverse_field_mapping_') or self._inverse_field_mapping_ is None:
            self._inverse_field_mapping_ = {value: key for key, value in self.field_mapping.items()}
        return self._inverse_field_mapping_

    def _column_likes_to_colnames(self, column_likes: List[ColumnLike]):
        if column_likes is None:
            return None

        feature_names = get_features_names(column_likes)
        column_names = list(map(lambda name: self._inverse_field_mapping.get(name, name), feature_names))
        return column_names



import pandas as pd
import pytest

import snax.data_sources.examples.csv
from snax.utils import frames_equal_up_to_row_ordering


@pytest.fixture
def nhl_data_source():
    return snax.data_sources.examples.csv.create_nhl_games()


def test_select_with_where(nhl_data_source):
    data = nhl_data_source.select(
        columns=['game_id', 'home_goals'],
        where_sql_query='game_id in (2016020045, 2017020812, 2015020314)'
    )
    expected_data = pd.DataFrame({'game_id': [2016020045, 2017020812, 2015020314], 'home_goals': [7, 3, 1]})
    assert frames_equal_up_to_row_ordering(data, expected_data)



import logging
from typing import List, Dict, Optional

import numpy as np
import pandas as pd
import sqlalchemy
from sqlalchemy import MetaData, Table, String, Integer, Float, Boolean
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.sql.type_api import TypeEngine

logger = logging.getLogger(__name__)


def ensure_table_exists(table: str, schema: str, engine: Engine):
    """Checks if schema.table exists in the Oracle DB and creates it if it doesn't"""
    query = f'SELECT * FROM {schema}.{table}'
    try:
        pd.read_sql(query, engine, chunksize=1)
        logger.info(f'Table {schema}.{table} exists')
    except Exception:
        query = f'CREATE TABLE {schema}.{table} (dummy int)'
        engine.execute(query)
        logger.info(f'Table {schema}.{table} created')


def drop_table(table: str, schema: str, engine: Engine):
    """Drops schema.table from the Oracle DB"""
    query = f'DROP TABLE {schema}.{table}'
    try:
        engine.execute(query)
        logger.info(f'Table {schema}.{table} dropped')
    except sqlalchemy.exc.DatabaseError as exception:
        if len(exception.args) > 0 and 'ORA-00942' in exception.args[0]:
            logger.debug(f'Can\'t drop table that doesn\'t exist: {exception.args[0]}')
        else:
            raise exception


def get_sqlalchemy_table(table: str, schema: str, engine: Engine) -> Table:
    return Table(table, MetaData(), autoload_with=engine, schema=schema)


def get_colnames(table: str, schema: str, engine: Engine) -> List[str]:
    """Returns a list of column names for schema.table"""
    table = get_sqlalchemy_table(table, schema, engine)
    return [col.name for col in table.columns]


def get_sqlalchemy_column_types(table: str, schema: str, engine: Engine) -> Dict[str, Optional[type]]:
    sqlalchemy_table = get_sqlalchemy_table(table, schema, engine)
    colname_to_type = dict()
    for column in sqlalchemy_table.columns:
        colname_to_type[column.name] = column.type
    return colname_to_type


def get_base_column_types(table: str, schema: str, engine: Engine) -> Dict[str, Optional[type]]:
    sqlalchemy_column_types = get_sqlalchemy_column_types(table, schema, engine)
    return {colname: sqlalchemy_column_type_to_base_type(column_type) for colname, column_type in
            sqlalchemy_column_types.items()}


def add_unique_constraint(key: List[str], table: str, schema: str, engine: Engine):
    constraint_name = schema.upper() + '_' + table.upper() + '_' + '_'.join([k.upper() for k in key]) + '_unique'
    sql = f'ALTER TABLE {schema}.{table} ADD CONSTRAINT {constraint_name} UNIQUE ({", ".join(key)})'
    try:
        engine.execute(sql)
    except DatabaseError as exception:
        if len(exception.args) > 0 and 'ORA-02261' in exception.args[0]:
            logger.debug(f'Unique constraint {constraint_name} already exists')
        else:
            raise exception


def get_oracle_type(column_type: type) -> str:
    if isinstance(column_type, int):
        return 'NUMBER'
    elif isinstance(column_type, float):
        return 'FLOAT'
    elif isinstance(column_type, str):
        return 'VARCHAR2'
    elif isinstance(column_type, bool):
        return 'NUMBER'
    else:
        return 'CLOB'


def add_columns(columns: List[str], data: pd.DataFrame, table: str, schema: str, engine: Engine):
    for column in columns:
        column_type = data[column].dtype
        oracle_type = get_oracle_type(column_type)
        sql = f'ALTER TABLE {schema}.{table} ADD {column} {oracle_type}'
        engine.execute(sql)
        logger.info(f'Added column {column} to table {schema}.{table}')


def upsert(key: List[str], columns: List[str], data: pd.DataFrame, table: str, schema: str, engine: Engine):
    tmp_table = f'{table}_tmp'
    table_column_types = get_sqlalchemy_column_types(table, schema, engine)
    data[key + columns].to_sql(
        tmp_table, engine, if_exists='replace', schema=schema, index=False, dtype=table_column_types)

    condition = ' and '.join([f'{table}.{key_} = {tmp_table}.{key_}' for key_ in key])
    sql = f'merge into {schema}.{table} using {schema}.{tmp_table} on ({condition}) ' \
          f'when matched then update set {", ".join([f"{table}.{column} = {tmp_table}.{column}" for column in columns])} ' \
          f'when not matched then insert ({", ".join(key + columns)}) values ({", ".join([f"{tmp_table}.{column}" for column in (key + columns)])})'

    with engine.begin() as conn:
        # See https://github.com/sqlalchemy/sqlalchemy/issues/5405 for details why we do not use just engine.execute()
        conn.execute(sql)

    drop_table(tmp_table, schema, engine)


def get_data_subset_in_db(data: pd.DataFrame, colnames: List[str], table: str, schema: str,
                          engine: Engine) -> pd.DataFrame:
    """Returns subset of data[colnames] that already exist in the Oracle DB"""
    colnames_separated_by_comma = ', '.join(colnames)
    value_tuples_separated_by_comma = pd_dataframe_to_comma_separated_tuples(data[colnames])
    data_in_db = pd.read_sql(
        sql=f'SELECT {colnames_separated_by_comma} FROM {schema}.{table} ' \
            f'WHERE ({colnames_separated_by_comma}) IN ({value_tuples_separated_by_comma})',
        con=engine
    )
    return data_in_db


# ----------------------------------------------------------------------------------------------------------------------

SQLALCHEMY_TO_PYTHON_TYPE = {
    String: str,
    Integer: int,
    Float: float,
    Boolean: int  # oracle does not have a boolean types
}


def sqlalchemy_column_type_to_base_type(column_type: TypeEngine) -> Optional[type]:
    for sqlalchemy_type, python_type in SQLALCHEMY_TO_PYTHON_TYPE.items():
        if issubclass(column_type.__class__, sqlalchemy_type):
            return python_type


def retype_dataframe(colname_to_type: Dict[str, Optional[type]], data: pd.DataFrame) -> pd.DataFrame:
    data = data.copy()
    for colname, coltype in colname_to_type.items():
        if colname in data and coltype is not None:
            data[colname] = data[colname].astype(coltype)
    return data


def pd_dataframe_to_comma_separated_tuples(data: pd.DataFrame) -> str:
    return ', '.join(list(data.apply(pd_series_to_comma_separated_tuple, axis=1)))


def escape_value(value: str) -> str:
    if pd.isna(value):
        return 'null'

    if isinstance(value, str):
        return f"'{value}'"
    elif isinstance(value, bool):
        return str(int(value))
    else:
        return str(value)


def pd_series_to_comma_separated_tuple(series: pd.Series) -> str:
    comma_joined = ', '.join([escape_value(item) for item in series])
    return f'({comma_joined})'


numpy_dtype_to_oracle_dict = {
    np.dtype('O'): 'CLOB',
    np.dtype('int32'): 'FLOAT',
    np.dtype('int64'): 'NUMBER',
    np.dtype('float32'): 'FLOAT',
    np.dtype('float64'): 'FLOAT',
    np.dtype('bool'): 'NUMBER'
}


def numpy_dtype_to_oracle(dtype: type) -> str:
    return numpy_dtype_to_oracle_dict.get(dtype, 'CLOB')


def add_column(column: str, dtype: type, table: str, schema: str, engine: Engine):
    oracle_dtype = numpy_dtype_to_oracle(dtype)
    sql = f'alter table {schema}.{table} add {column} {oracle_dtype}'
    engine.execute(sql)


def ensure_columns_exist(columns: List[str], dtypes: Dict[str, type], table: str, schema: str, engine: Engine):
    columns_in_db = get_colnames(table, schema, engine)
    for column in columns:
        if column not in columns_in_db:
            add_column(column, dtypes[column], table, schema, engine)



from typing import Optional, Dict, List, Any

import pandas as pd

from snax.data_sources.data_source_base import DataSourceBase


def escape(value: Any) -> str:
    if isinstance(value, str):
        return f"'{value}'"
    return value


class InMemoryDataSource(DataSourceBase):
    def __init__(self, name: str, data: Optional[pd.DataFrame] = None, field_mapping: Optional[Dict[str, str]] = None,
                 tags: Optional[Dict] = None):
        super().__init__(name=name, field_mapping=field_mapping, tags=tags)
        self._data = data

    def _select(self, columns: Optional[List[str]] = None, where_sql_query: Optional[str] = None) -> pd.DataFrame:
        data_subset = self._data.copy()
        if where_sql_query is not None:
            data_subset = data_subset.query(where_sql_query)

        if columns is not None:
            return data_subset.loc[:, columns]
        else:
            return data_subset

    def _insert(self, key: List[str], columns: List[str], data: pd.DataFrame, if_exists: str = 'error'):
        self._ensure_key_in_data(key)
        data_to_insert = data[key + columns].copy()

        data_index = pd.MultiIndex.from_frame(self._data[key])
        insert_index = pd.MultiIndex.from_frame(data_to_insert[key])

        data_indices_of_inserted_rows = data_index.get_indexer_for(insert_index)

        if if_exists == 'error':
            existing_data_colnames = list(self._data.columns)
            inserted_data_have_some_original_colnames = any(colname in existing_data_colnames for colname in columns)
            inserted_data_have_some_original_keys = any(data_indices_of_inserted_rows != -1)
            if inserted_data_have_some_original_colnames and inserted_data_have_some_original_keys:
                raise ValueError('Some of the inserted data already exists in the data source')

        new_columns = [colname for colname in data_to_insert.columns if colname not in self._data.columns]
        for original_index_of_inserted_row, (_, inserted_row) \
                in zip(data_indices_of_inserted_rows, data_to_insert.iterrows()):
            # TODO: Optimize speed by splitting on indices with -1 and not -1
            if original_index_of_inserted_row == -1:
                self._data = self._data.append(inserted_row)
            else:
                if if_exists == 'replace' or if_exists == 'error':
                    self._data.loc[original_index_of_inserted_row, columns] = inserted_row
                elif if_exists == 'ignore':
                    self._data.loc[original_index_of_inserted_row, new_columns] = new_columns
                else:
                    raise ValueError(f'Unknown if_exists value: {if_exists}')

    def _where_sql_query_from_key_values(self, key: List[str], key_values: pd.DataFrame) -> str:
        row_to_string = lambda row: '(' + ' and '.join(f'{key_}=={escape(row[key_])}' for key_ in key) + ')'
        query = ' or '.join(list(key_values.apply(row_to_string, axis=1)))
        return query

    def _ensure_key_in_data(self, key: List[str]):
        for key_ in key:
            if key_ not in self._data:
                self._data[key_] = None



from pathlib import Path

import snax.example_feature_repos.winequality_feature_repo as winequality_feature_repo
from snax.data_sources.csv_data_source import CsvDataSource
from snax.feature import Feature
from snax.feature_view import FeatureView
from snax.repo_contents import parse_repo, DUMMY_ENTITY
from snax.utils import copy_to_temp
from snax.value_type import Float


def test_parse_repo():
    # Copy to arbitrary location to simulate real-world use-case
    original_wine_quality_repo_path = Path(winequality_feature_repo.__file__).parent
    wine_quality_repo_path = copy_to_temp(original_wine_quality_repo_path)

    repo_contents = parse_repo(repo_path=wine_quality_repo_path)

    assert len(repo_contents.entities) == 1
    assert len(repo_contents.data_sources) == 1
    assert len(repo_contents.feature_views) == 1

    expected_entity = DUMMY_ENTITY
    expected_data_source = CsvDataSource(
        name='winequality_csv',
        csv_file_path=None
    )
    expected_feature_view = FeatureView(
        name='winequality_acidity_csv',
        entities=None,
        features=[
            Feature('volatile_acidity', Float),
            Feature('fixed_acidity', Float),
            Feature('citric_acidity', Float),
            Feature('ph', Float)
        ],
        source=expected_data_source
    )

    assert repo_contents.entities[0] == expected_entity
    assert repo_contents.data_sources[0] == expected_data_source
    assert repo_contents.feature_views[0] == expected_feature_view



from typing import Union, List

from snax.entity import Entity
from snax.feature import Feature

ColumnLike = Union[Entity, Feature, str]


def get_feature_names(column_like: ColumnLike) -> List[str]:
    if isinstance(column_like, Entity):
        return column_like.join_keys
    elif isinstance(column_like, Feature):
        return [column_like.name]
    elif isinstance(column_like, str):
        return [column_like]
    else:
        raise TypeError(f'Unsupported type {type(column_like)}')


def get_features_names(column_likes: List[ColumnLike]) -> List[str]:
    column_likes_nested = [get_feature_names(column_like) for column_like in column_likes]
    return sum(column_likes_nested, [])



import os
import shutil
import tempfile
from pathlib import Path

import pandas as pd


def copy_to_temp(path: str) -> str:
    """
    Copy a file or a directory to a temporary location and return the path to that location.
    Args:
        path: Path of the file / directory to be copied

    Returns:
        Path to the copied file / directory
    """
    snax_tmp_path = Path(tempfile.gettempdir()) / 'snax_tmp'
    os.makedirs(snax_tmp_path, exist_ok=True)

    target_path = snax_tmp_path / Path(path).name

    if os.path.isdir(path):
        if os.path.exists(target_path):
            shutil.rmtree(target_path)

        return shutil.copytree(path, target_path)
    else:
        if os.path.exists(target_path):
            os.remove(target_path)

        return shutil.copy(path, target_path)


def frames_equal_up_to_row_ordering(df1: pd.DataFrame, df2: pd.DataFrame) -> bool:
    if (df1.columns != df2.columns).any():
        return False
    merged = df1.merge(df2, on=df1.columns.tolist(), how='outer', indicator=True)
    return (merged['_merge'] == 'both').all()



from typing import Dict, Optional

from snax.value_type import ValueType


class Feature:
    """
    A property that we measure / observe

    Args:
        name: Name of the feature
        dtype: Data type of the feature
        tags: Tags for the feature
    """

    def __init__(self, name: str, dtype: ValueType, tags: Optional[Dict[str, str]] = None):
        self._name = name
        self._dtype = dtype
        self._tags = tags or dict()

    def __repr__(self):
        return f'Feature(name={self.name}, dtype={self.dtype})'

    def __eq__(self, other):
        if not isinstance(other, Feature):
            return False

        if (self.name != other.name) or (self.dtype != other.dtype):
            return False

        return True

    def __hash__(self):
        return hash((self.name, self.dtype))

    @property
    def name(self) -> str:
        return self._name

    @property
    def dtype(self) -> ValueType:
        return self._dtype

    @property
    def tags(self) -> Dict[str, str]:
        return self._tags



from tempfile import NamedTemporaryFile

from snax.data_sources.csv_data_source import CsvDataSource
from snax.example_feature_repos.sports_feature_repo.nhl_games import data_path as original_nhl_data_path
from snax.example_feature_repos.users_with_nas_feature_repo.users_with_nas import \
    data_path as original_users_with_na_data_path
from snax.utils import copy_to_temp


def create_nhl_games() -> CsvDataSource:
    return CsvDataSource(
        name='nhl_games_csv',
        csv_file_path=copy_to_temp(original_nhl_data_path)
    )


def create_users_with_nas() -> CsvDataSource:
    return CsvDataSource(
        name='users_with_nas_csv',
        csv_file_path=copy_to_temp(original_users_with_na_data_path)
    )


def create_users_with_nas_field_mapping() -> CsvDataSource:
    return CsvDataSource(
        name='users_with_nas_csv',
        csv_file_path=copy_to_temp(original_users_with_na_data_path),
        field_mapping={
            'is_subscribed': 'issubscribed',
            'timestamp': 'time_stamp',
        }
    )


def create_empty_data_source() -> CsvDataSource:
    tmp_file = NamedTemporaryFile(suffix='.csv')
    return CsvDataSource(
        name='empty_csv',
        csv_file_path=tmp_file.name
    )



from typing import List, Optional, Dict

import pandas as pd

from snax.data_sources.data_source_base import DataSourceBase
from snax.entity import Entity
from snax.feature import Feature
from snax.type_casting import cast_to_feature_types


class FeatureView:
    """
    Group of features and entities from a single data source that create logical group of data

    Args:
        name: Name of the feature view
        entities: List of entities that are part of this feature view
        features: List of features that are part of this feature view
        source: Data source that this feature view is based on
        tags: Tags for this feature view
    """

    def __init__(self, name: str, entities: Optional[List[Entity]], features: Optional[List[Feature]],
                 source: DataSourceBase, tags: Optional[Dict[str, str]] = None):
        self._name = name
        self._entities = entities
        self._features = features
        self._source = source
        self._tags = tags or dict()

    def __repr__(self):
        return f'FeatureView(name={self.name})'

    def __eq__(self, other):
        if not isinstance(other, FeatureView):
            return False

        if self.name != other.name:
            return False

        if ((self.entities is None) and (other.entities is not None)) or \
                ((self.entities is not None) and (other.entities is None)):
            return False

        if self.entities is not None:
            for entity in self.entities:
                if entity not in other.entities:
                    return False

        if ((self.features is None) and (other.features is not None)) or \
                ((self.features is not None) and (other.features is None)):
            return False

        if self.features is not None:
            for feature in self.features:
                if feature not in other.features:
                    return False

        if self.source != other.source:
            return False

        return True

    def __hash__(self):
        return hash((self.name, self.entities, self.features, self.source))

    @property
    def name(self) -> str:
        return self._name

    @property
    def entities(self) -> Optional[List[str]]:
        return self._entities

    @property
    def features(self) -> List[Feature]:
        return self._features

    @property
    def source(self) -> DataSourceBase:
        return self._source

    @property
    def tags(self) -> Dict[str, str]:
        return self._tags

    def get_entity(self, entity_name: str) -> Entity:
        for entity in self.entities:
            if entity.name == entity_name:
                return entity
        raise ValueError(f'Entity {entity_name} not found in feature view {self.name}')

    def get_feature(self, feature_name: str) -> Feature:
        for feature in self.features:
            if feature.name == feature_name:
                return feature
        raise ValueError(f'Feature {feature_name} not found in feature view {self.name}')

    def add_features_to_dataframe(self, dataframe: pd.DataFrame, feature_names: List[str],
                                  entity_name: Optional[str] = None) -> pd.DataFrame:
        if entity_name is None:
            raise NotImplementedError('Joins without entity not supported yet. ')

        entity = self.get_entity(entity_name)

        feature_values = self.source.select(
            columns=[entity] + feature_names,
            key=[entity],
            key_values=dataframe[entity.join_keys]
        )

        feature_values = cast_to_feature_types(
            dataframe=feature_values,
            features=[self.get_feature(feature_name) for feature_name in feature_names]
        )

        dataframe = dataframe.merge(feature_values, on=entity.join_keys, how='left')
        return dataframe



from typing import Union
from unittest import skip

import pandas as pd
import pytest
from _pytest.outcomes import Skipped
from pandas.testing import assert_frame_equal

from snax.data_sources.data_source_base import DataSourceBase
from snax.data_sources.examples import csv, in_memory, oracle
from snax.data_sources.oracle_data_source import OracleDataSource
from snax.entity import Entity
from snax.feature import Feature
from snax.utils import frames_equal_up_to_row_ordering
from snax.value_type import Int, String, Bool, Timestamp

_data_source_backend_to_examples_module = {
    'csv': csv,
    'in-memory': in_memory,
    'oracle': oracle
}

_data_source_backends = ['csv', 'in-memory', 'oracle']


def _handle_unavailable_datasource(data_source_backend: str, data_source: DataSourceBase) -> Union[
    Skipped, DataSourceBase]:
    if data_source_backend == 'oracle' and data_source is None:
        return pytest.skip('Oracle backend is not available')
    else:
        return data_source


@pytest.fixture(params=_data_source_backends)
def nhl_data_source(request):
    module = _data_source_backend_to_examples_module[request.param]
    data_source = module.create_nhl_games()
    return _handle_unavailable_datasource(request.param, data_source)


@pytest.fixture(params=_data_source_backends)
def users_with_nas_data_source(request):
    module = _data_source_backend_to_examples_module[request.param]
    data_source = module.create_users_with_nas()
    return _handle_unavailable_datasource(request.param, data_source)


@pytest.fixture(params=_data_source_backends)
def users_with_nas_field_mapping_data_source(request):
    module = _data_source_backend_to_examples_module[request.param]
    data_source = module.create_users_with_nas_field_mapping()
    return _handle_unavailable_datasource(request.param, data_source)


@pytest.fixture(params=_data_source_backends)
def empty_data_source(request):
    module = _data_source_backend_to_examples_module[request.param]
    data_source = module.create_empty_data_source()
    return _handle_unavailable_datasource(request.param, data_source)


def test_select_single_column_by_string(nhl_data_source):
    data = nhl_data_source.select(['game_id'])
    first_few_sorted_game_ids = sorted(list(data['game_id']))[:5]
    expected_first_few_sorted_game_ids = [2015020007, 2015020061, 2015020079, 2015020086, 2015020165]
    assert first_few_sorted_game_ids == expected_first_few_sorted_game_ids


def test_select_single_column_by_feature(nhl_data_source):
    data = nhl_data_source.select([Feature('game_id', Int)])
    first_few_sorted_game_ids = sorted(list(data['game_id']))[:5]
    expected_first_few_sorted_game_ids = [2015020007, 2015020061, 2015020079, 2015020086, 2015020165]
    assert first_few_sorted_game_ids == expected_first_few_sorted_game_ids


def test_select_single_column_by_entity(nhl_data_source):
    data = nhl_data_source.select([Entity('game', join_keys=['game_id'])])
    first_few_sorted_game_ids = sorted(list(data['game_id']))[:5]
    expected_first_few_sorted_game_ids = [2015020007, 2015020061, 2015020079, 2015020086, 2015020165]
    assert first_few_sorted_game_ids == expected_first_few_sorted_game_ids


def test_select_multiple_columns(nhl_data_source):
    data = nhl_data_source.select(
        columns=[
            Entity('game', join_keys=['game_id']),
            Feature('game_id', Int),
            'game_id',
            Feature('away_team_id', Int)
        ]
    )
    expected_data_columns = ['game_id', 'game_id', 'game_id', 'away_team_id']
    assert list(data.columns) == expected_data_columns


def test_select_string_column_with_missings(users_with_nas_data_source):
    data = users_with_nas_data_source.select(['first_name'])
    assert isinstance(data, pd.DataFrame)


def test_select_int_column_with_missings(users_with_nas_data_source):
    data = users_with_nas_data_source.select(['children'])
    assert isinstance(data, pd.DataFrame)


def test_select_float_column_with_missings(users_with_nas_data_source):
    data = users_with_nas_data_source.select(['age'])
    assert isinstance(data, pd.DataFrame)


def test_select_timestamp_column_with_missings(users_with_nas_data_source):
    data = users_with_nas_data_source.select(['timestamp'])
    assert isinstance(data, pd.DataFrame)


def test_select_all_columns(users_with_nas_data_source):
    data = users_with_nas_data_source.select()
    assert isinstance(data, pd.DataFrame)


def test_select_by_key_values_single_int_index(nhl_data_source):
    key_values = pd.DataFrame({'game_id': [2016020045, 2017020812, 2015020314]})
    selected_data = nhl_data_source.select(
        columns=['game_id', 'season', 'type'], key=['game_id'], key_values=key_values)
    expected_selected_data = pd.DataFrame({
        'game_id': [2016020045, 2017020812, 2015020314],
        'season': [20162017, 20172018, 20152016],
        'type': ['R', 'R', 'R']
    })
    assert frames_equal_up_to_row_ordering(selected_data, expected_selected_data)


def test_select_by_key_values_single_int_index(nhl_data_source):
    if isinstance(nhl_data_source, OracleDataSource):
        # TODO: Implement corresponding behavior for Oracle
        return skip('Handling case-sensitive column names not yet implemented for Oracle data source')

    key_values = pd.DataFrame({'game_id': [2016020045, 2017020812, 2015020314]})
    selected_data = nhl_data_source.select(
        columns=['game_id', 'date_time_GMT'], key=['game_id'], key_values=key_values)
    expected_selected_data = pd.DataFrame({
        'game_id': [2016020045, 2017020812, 2015020314],
        'date_time_GMT': ['2016-10-19T00:30:00Z', '2018-02-07T00:00:00Z', '2015-11-24T01:00:00Z']
    })
    assert frames_equal_up_to_row_ordering(selected_data, expected_selected_data)


def test_select_by_key_values_single_string_index(users_with_nas_data_source):
    key_values = pd.DataFrame({'string_id': ['a', 'b', 'c']})
    selected_data = users_with_nas_data_source.select(
        columns=['string_id', 'id', 'first_name'], key=['string_id'], key_values=key_values)
    expected_selected_data = pd.DataFrame({
        'string_id': ['a', 'b', 'c'],
        'id': [1, 2, 3],
        'first_name': ['Cirillo', 'Codi', 'Marion']
    })
    assert frames_equal_up_to_row_ordering(selected_data, expected_selected_data)


def test_select_by_key_values_string_int_index(users_with_nas_data_source):
    key_values = pd.DataFrame({'string_id': ['a', 'b', 'c'], 'id': [1, 2, 3]})
    selected_data = users_with_nas_data_source.select(
        columns=['string_id', 'id', 'first_name'], key=['string_id', 'id'], key_values=key_values)
    expected_selected_data = pd.DataFrame({
        'string_id': ['a', 'b', 'c'],
        'id': [1, 2, 3],
        'first_name': ['Cirillo', 'Codi', 'Marion']
    })
    assert frames_equal_up_to_row_ordering(selected_data, expected_selected_data)


def test_insert_validates_argument(users_with_nas_data_source):
    with pytest.raises(ValueError) as exception_info:
        users_with_nas_data_source.insert(['id'], ['first_name'], pd.DataFrame(), if_exists='foobar')


def test_insert_full_new_column(users_with_nas_data_source):
    first_names_uppercase = [
        'CIRILLO', 'CODI', 'MARION', 'GERMAINE', None, 'MELLONEY', 'MADEL', None, 'ROCKWELL', 'TRUMAINE']

    users_with_nas_data_source.insert(
        key=[Entity('user', join_keys=['id'])],
        columns=[Feature('first_name_uppercase', String)],
        data=pd.DataFrame({
            'id': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'first_name_uppercase': first_names_uppercase
        })
    )
    data = users_with_nas_data_source.select(['first_name_uppercase'])
    assert set(first_names_uppercase) == set(list(data['first_name_uppercase']))


def test_insert_partial_new_column(users_with_nas_data_source):
    first_names_uppercase = ['MARION', None, 'TRUMAINE']

    users_with_nas_data_source.insert(
        key=[Entity('user', join_keys=['id'])],
        columns=[Feature('first_name_uppercase', String)],
        data=pd.DataFrame({
            'id': [3, 5, 10],
            'first_name_uppercase': first_names_uppercase
        })
    )
    data = users_with_nas_data_source.select(['id', 'first_name_uppercase'])
    assert (~data['first_name_uppercase'].isna()).sum() == 2
    data_updated = data[data['id'].isin([3, 5, 10])]
    data_updated.sort_values(by='id', inplace=True)
    assert first_names_uppercase == list(data_updated['first_name_uppercase'])


def test_insert_full_new_row(users_with_nas_data_source):
    new_data = pd.DataFrame({
        'id': [0, 11],
        'first_name': ['CIRILLO', 'CODI'],
        'last_name': ['SEABORN', 'PANDAS'],
        'gender': ['Male', 'Female'],
        'timestamp': ['2022-08-17T10:41:25', '2022-08-16T15:45:31'],
        'age': [None, 56],
        'is_subscribed': [False, True],
        'children': [2, None],
        'string_id': ['x', 'y']
    })
    users_with_nas_data_source.insert(
        key=[Entity('user', join_keys=['id'])],
        columns=[Feature('first_name', dtype=String),
                 'last_name', 'gender', 'timestamp', 'age', 'is_subscribed', 'children', 'string_id'],
        data=new_data
    )
    retrieved_data = users_with_nas_data_source.select()
    filtered_retrieved_data = retrieved_data[retrieved_data['id'].isin([0, 11])]
    assert_frame_equal(new_data.reset_index(drop=True), filtered_retrieved_data.reset_index(drop=True),
                       check_dtype=False)


def test_insert_partial_new_row(users_with_nas_data_source):
    data = pd.DataFrame({
        'id': [0, 11],
        'gender': ['Male', 'Female'],
    })
    users_with_nas_data_source.insert(
        key=[Entity('user', join_keys=['id'])],
        columns=['gender'],
        data=data,
        if_exists='error'
    )
    retrieved_data = users_with_nas_data_source.select([Entity('user', join_keys=['id']), Feature('gender', String)])
    filtered_retrieved_data = retrieved_data[retrieved_data['id'].isin([0, 11])]
    assert_frame_equal(data.reset_index(drop=True), filtered_retrieved_data[['id', 'gender']].reset_index(drop=True))


def test_insert_existing_data_replace(users_with_nas_data_source):
    data = pd.DataFrame({
        'id': [4, 5, 6],
        'first_name': [None, 'Smith', 'Smith']
    })
    users_with_nas_data_source.insert(
        key=['id'],
        columns=['first_name'],
        data=data,
        if_exists='replace'
    )
    retrieved_data = users_with_nas_data_source.select(['id', 'first_name'])
    retrieved_data.sort_values(by='id', inplace=True)
    filtered_retrieved_data = retrieved_data[retrieved_data['id'].isin([4, 5, 6])]
    assert_frame_equal(data.reset_index(drop=True), filtered_retrieved_data.reset_index(drop=True))


def test_insert_multikey_int_int(nhl_data_source):
    data = pd.DataFrame({
        'game_id': [2016020045, 2017020812, 2015020314],
        'season': [20162017, 20172018, 20152016],
        'home_goals': [3, 2, 1],
    })
    nhl_data_source.insert(
        key=['game_id', 'season'],
        columns=['home_goals'],
        data=data,
        if_exists='replace'
    )
    retrieved_data = nhl_data_source.select(['game_id', 'season', 'home_goals'])
    filtered_retrieved_data = retrieved_data[retrieved_data['game_id'].isin([2016020045, 2017020812, 2015020314])]
    assert frames_equal_up_to_row_ordering(data, filtered_retrieved_data)


def test_insert_multikey_int_str(nhl_data_source):
    data = pd.DataFrame({
        'game_id': [2016020045, 2017020812, 2015020314],
        'venue': ['United Center', 'KeyBank Center', 'foo bar'],
        'home_goals': [0, 1, 2]
    })
    nhl_data_source.insert(
        key=['game_id', 'venue'],
        columns=['home_goals'],
        data=data,
        if_exists='replace'
    )
    retrieved_data = nhl_data_source.select(['game_id', 'venue', 'home_goals'])
    filtered_retrieved_data = retrieved_data[retrieved_data['game_id'].isin([2016020045, 2017020812, 2015020314])]
    expected_filtered_retrieved_data = pd.DataFrame([
        {'game_id': 2016020045, 'venue': 'United Center', 'home_goals': 0},
        {'game_id': 2017020812, 'venue': 'KeyBank Center', 'home_goals': 1},
        {'game_id': 2015020314, 'venue': 'MTS Centre', 'home_goals': 1},
        {'game_id': 2015020314, 'venue': 'foo bar', 'home_goals': 2}
    ])
    assert frames_equal_up_to_row_ordering(filtered_retrieved_data, expected_filtered_retrieved_data)


def test_insert_existing_data_error(users_with_nas_data_source):
    with pytest.raises(ValueError) as exception_info:
        data = pd.DataFrame({'id': [4], 'first_name': [None]})
        users_with_nas_data_source.insert(key=['id'], columns=['first_name'], data=data, if_exists='error')


def test_insert_existing_data_ignore(users_with_nas_data_source):
    data = pd.DataFrame({'id': [4], 'first_name': [None]})
    users_with_nas_data_source.insert(key=['id'], columns=['first_name'], data=data, if_exists='ignore')
    retrieved_data = users_with_nas_data_source.select(['id', 'first_name'])

    filtered_retrieved_data = retrieved_data[retrieved_data['id'] == 4]
    assert str(filtered_retrieved_data['first_name'].iloc[0]) == 'Germaine'


def test_select_from_datasource_with_field_mapping(users_with_nas_field_mapping_data_source):
    data = users_with_nas_field_mapping_data_source.select(['time_stamp', Feature('issubscribed', Bool)])
    assert list(data.columns) == ['time_stamp', 'issubscribed']


def test_insert_to_datasource_with_field_mapping(users_with_nas_field_mapping_data_source):
    data = pd.DataFrame({
        'id': [10, 11],
        'time_stamp': ['2030-08-17T10:41:25', '2030-08-16T15:45:31'],
        'issubscribed': [True, True]
    })
    users_with_nas_field_mapping_data_source.insert(
        key=[Entity('user', join_keys=['id'])],
        columns=[Feature('time_stamp', Timestamp), Feature('issubscribed', Bool)],
        data=data,
        if_exists='replace'
    )
    retrieved_data = users_with_nas_field_mapping_data_source.select(
        ['id', 'time_stamp', Feature('issubscribed', Bool)])
    retrieved_data = retrieved_data[retrieved_data['id'].isin([10, 11])]
    retrieved_data.sort_values(by='id', inplace=True)
    assert_frame_equal(data.reset_index(drop=True), retrieved_data.reset_index(drop=True), check_dtype=False)


def test_insert_into_empty_datasource(empty_data_source):
    data = pd.DataFrame({
        'id': [1, 2, 3],
        'first_name': ['John', 'Jane', 'Mary'],
        'last_name': ['Doe', 'Doe', 'Doe']
    })
    empty_data_source.insert(
        key=['id'],
        columns=['first_name', 'last_name'],
        data=data,
        if_exists='replace'
    )
    retrieved_data = empty_data_source.select(['id', 'first_name', 'last_name'])
    assert frames_equal_up_to_row_ordering(data, retrieved_data)



from pathlib import Path

from snax.data_sources.csv_data_source import CsvDataSource
from snax.feature import Feature
from snax.feature_view import FeatureView
from snax.value_type import Float

data_path = Path(__file__).parent.parent.parent / 'data' / 'winequality.csv'

# Define DataSource Objects --------------------------------------------------------------------------------------------
winequality_csv_data_source = CsvDataSource(
    name='winequality_csv',
    csv_file_path=data_path,
    separator=',',
    field_mapping={
        'citric_acid': 'citric_acidity',
        'pH': 'ph'
    },
    tags={
        'owner': 'john.doe@gmail.com'
    }
)

# Define FeatureView Objects -------------------------------------------------------------------------------------------
winequality_csv_acidity_feature_view = FeatureView(
    name='winequality_acidity_csv',
    entities=None,
    features=[
        Feature('volatile_acidity', Float),
        Feature('fixed_acidity', Float),
        Feature('citric_acidity', Float),
        Feature('ph', Float)
    ],
    source=winequality_csv_data_source
)



import pandas as pd

from snax.data_sources.in_memory_data_source import InMemoryDataSource
from snax.example_feature_repos.sports_feature_repo.nhl_games import data_path as original_nhl_data_path
from snax.example_feature_repos.users_with_nas_feature_repo.users_with_nas import \
    data_path as original_users_with_na_data_path


def create_nhl_games() -> InMemoryDataSource:
    return InMemoryDataSource(
        name='nhl_games_in_memory',
        data=pd.read_csv(original_nhl_data_path)
    )


def create_users_with_nas() -> InMemoryDataSource:
    return InMemoryDataSource(
        name='users_with_nas_in_memory',
        data=pd.read_csv(original_users_with_na_data_path)
    )


def create_users_with_nas_field_mapping() -> InMemoryDataSource:
    return InMemoryDataSource(
        name='users_with_nas_in_memory',
        data=pd.read_csv(original_users_with_na_data_path),
        field_mapping={
            'is_subscribed': 'issubscribed',
            'timestamp': 'time_stamp',
        }
    )


def create_empty_data_source() -> InMemoryDataSource:
    return InMemoryDataSource(
        name='empty_in_memory',
        data=pd.DataFrame()
    )



from datetime import datetime

import pandas as pd
import pytest

from snax.feature import Feature
from snax.type_casting import guess_timestamp_format, cast_to_feature_types, cast_to_feature_type
from snax.value_type import ValueType


def test_cast_to_feature_types():
    data = pd.DataFrame({
        'timestamp': ['2020-01-01', '2020-01-02', '2020-01-03'],
        'bool': ['true', 'false', 'true'],
        'float': ['1.0', '2.0', '3.0'],
        'int': ['1', '2', '3'],
        'string': ['a', 'b', 'c']
    })
    cast_data = cast_to_feature_types(
        dataframe=data,
        features=[
            Feature(name='timestamp', dtype=ValueType.TIMESTAMP),
            Feature(name='bool', dtype=ValueType.BOOL),
            Feature(name='float', dtype=ValueType.FLOAT),
            Feature(name='int', dtype=ValueType.INT),
            Feature(name='string', dtype=ValueType.STRING)
        ]
    )

    assert cast_data['timestamp'].iloc[0] == datetime(2020, 1, 1)
    assert cast_data['bool'].iloc[0] == True
    assert cast_data['float'].iloc[0] == 1.0
    assert cast_data['int'].iloc[0] == 1
    assert cast_data['string'].iloc[0] == 'a'


def test_guess_timestamp_format():
    assert guess_timestamp_format(pd.Series(['2020-01-01', '2020-01-02', '2020-01-03'])) == '%Y-%m-%d'
    assert guess_timestamp_format(pd.Series(['01.01.2020', '01.02.2020', '01.03.2020'])) == '%d.%m.%Y'
    assert guess_timestamp_format(pd.Series(['01:01:17', '01:01:18', '01:01:19'])) == '%H:%M:%S'
    assert guess_timestamp_format(pd.Series(['01:01:17.123', '01:01:18.123', '01:01:19.123'])) == '%H:%M:%S.%f'
    assert guess_timestamp_format(
        pd.Series(['2020-01-02 01:01:17', '2020-01-02 01:01:18', '2020-01-02 01:01:19'])) == '%Y-%m-%d %H:%M:%S'
    assert guess_timestamp_format(pd.Series(
        ['2020-12-02T01:01:17.123', '2020-12-02T01:01:18.123', '2020-12-02T01:01:19.123'])) == '%Y-%m-%dT%H:%M:%S.%f'


RAW_SERIES__FEATURE_TYPE__CAST_SERIES = [
    (
        pd.Series(['2020-01-01', '2020-01-02', None]),
        ValueType.TIMESTAMP,
        pd.Series([datetime(2020, 1, 1), datetime(2020, 1, 2), None])
    ),
    (
        pd.Series(['2020-01-01T01:01:17', '2020-01-02T01:01:18', None]),
        ValueType.TIMESTAMP,
        pd.Series([datetime(2020, 1, 1, 1, 1, 17), datetime(2020, 1, 2, 1, 1, 18), None])
    ),
    (
        pd.Series(['True', 'F', '0.0', None]),
        ValueType.BOOL,
        pd.Series([True, False, False, None])
    ),
    (
        pd.Series(['1.0', '2.0', '3.0', None]),
        ValueType.FLOAT,
        pd.Series([1.0, 2.0, 3.0, None])
    ),
    (
        pd.Series(['1', '2', '3', None]),
        ValueType.INT,
        pd.Series([1, 2, 3, None])
    ),
    (
        pd.Series(['a', 'b', 'c', None]),
        ValueType.STRING,
        pd.Series(['a', 'b', 'c', None])
    ),
    (
        pd.Series([
            '["2020-01-03H15:24:35", "2022-01-04H15:26:35", "2020-01-05H07:24:16", null]',
            '["2023-01-03H15:24:35", "2024-01-04H15:26:35"]',
        ]),
        ValueType.TIMESTAMP_LIST,
        pd.Series([
            [datetime(2020, 1, 3, 15, 24, 35), datetime(2022, 1, 4, 15, 26, 35), datetime(2020, 1, 5, 7, 24, 16), None],
            [datetime(2023, 1, 3, 15, 24, 35), datetime(2024, 1, 4, 15, 26, 35)]
        ])
    ),
    (
        pd.Series([
            '[true, false, true]',
            '[false, null, false]',
            '[true, false, null]'
        ]),
        ValueType.BOOL_LIST,
        pd.Series([
            [True, False, True],
            [False, None, False],
            [True, False, None]
        ])
    ),
    (
        pd.Series([
            '[1.0, 2.0, 3.0]',
            '[2.0, null, 4.0]',
            '[3.0, 4.0, null]'
        ]),
        ValueType.FLOAT_LIST,
        pd.Series([
            [1.0, 2.0, 3.0],
            [2.0, None, 4.0],
            [3.0, 4.0, None]
        ])
    ),
    (
        pd.Series([
            '[1, 2, 3]',
            '[2, 3, null]',
            '[3, null, 5]']
        ),
        ValueType.INT_LIST,
        pd.Series([
            [1, 2, 3],
            [2, 3, None],
            [3, None, 5]
        ])
    ),
    (
        pd.Series([
            '["a", "b", "c"]',
            '["b", "c", null]',
            '["c", "d", "e"]'
        ]),
        ValueType.STRING_LIST,
        pd.Series([
            ['a', 'b', 'c'],
            ['b', 'c', None],
            ['c', 'd', 'e']
        ])
    ),
]


@pytest.mark.parametrize('raw_series,feature_type,expected_cast_series', RAW_SERIES__FEATURE_TYPE__CAST_SERIES)
def test_cast_to_feature_type(raw_series, feature_type, expected_cast_series):
    cast_series = cast_to_feature_type(raw_series, feature_type=feature_type)
    assert cast_series.equals(expected_cast_series)



from setuptools import setup, find_packages

setup(
    name='snax',
    version='0.1.0',
    author='Josef Ondrej',
    author_email='josef.ondrej@outlook.com',
    description='The truly lightweight feature store',
    long_description='The truly lightweight feature store',
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={'snax': ['data/*.csv']},
    install_requires=[
        'pandas>=1.4.0',
        'sqlalchemy>=1.4.0'
    ]
)



from snax.column_like import get_feature_names
from snax.entity import Entity
from snax.feature import Feature
from snax.value_type import Int


def test_get_colnames_entity():
    assert get_feature_names(Entity('game', ['home_team_id', 'away_team_id'])) == ['home_team_id', 'away_team_id']


def test_get_colnames_feature():
    assert get_feature_names(Feature('home_team_id', Int)) == ['home_team_id']


def test_get_colnames_str():
    assert get_feature_names('home_team_id') == ['home_team_id']



from typing import List, Optional, Dict

import pandas as pd

from snax.data_sources.data_source_base import DataSourceBase
from snax.entity import Entity
from snax.feature_view import FeatureView
from snax.repo_contents import parse_repo, RepoContents


def group_features(features: List[str]) -> Dict[str, List[str]]:
    feature_dict = {}
    for feature in features:
        view_name, feature_name = feature.split(':')  # TODO: Define the splitter somewhere
        if view_name not in feature_dict:
            feature_dict[view_name] = []
        feature_dict[view_name].append(feature_name)

    return feature_dict


class FeatureStore:
    """
    Collection of various feature views
    """

    def __init__(self, repo_path: str):
        self._repo_path = repo_path
        self._repo_contents: RepoContents = parse_repo(repo_path)

    @property
    def repo_path(self) -> str:
        return self._repo_path

    def add_features_to_dataframe(self, dataframe: pd.DataFrame, feature_names: List[str],
                                  entity_name: Optional[str] = None) -> pd.DataFrame:
        """
        Retrieve features by their full name (feature_view_name:feature_name) and add them to the dataframe

        Args:
            dataframe: Entity's key values dataframe to add features to
            feature_names: List of full feature names to add to the dataframe in the format view_name:feature_name
            entity_name: Optional entity name to specify what columns to use for identifying the entity in the dataframe
                if it contains more columns than are required to identify the entity
        """
        feature_groups = group_features(feature_names)
        for view_name, feature_names in feature_groups.items():
            view = self.get_feature_view(view_name)
            dataframe = view.add_features_to_dataframe(dataframe, feature_names, entity_name)

        return dataframe

    def list_feature_views(self) -> List[FeatureView]:
        return self._repo_contents.feature_views

    def get_feature_view(self, name: str) -> FeatureView:
        return self._repo_contents.get_feature_view(name)

    def list_entities(self) -> List[Entity]:
        return self._repo_contents.entities

    def get_entity(self, name: str) -> Entity:
        return self._repo_contents.get_entity(name)

    def list_data_sources(self) -> List[DataSourceBase]:
        return self._repo_contents.data_sources

    def get_data_source(self, name: str) -> DataSourceBase:
        return self._repo_contents.get_data_source(name)



from pathlib import Path

from snax.data_sources.csv_data_source import CsvDataSource
from snax.entity import Entity
from snax.feature import Feature
from snax.feature_view import FeatureView
from snax.value_type import Int, String, Timestamp, Float, Bool

data_path = Path(__file__).parent.parent.parent / 'data' / 'users_with_nas.csv'

# Define DataSource Objects --------------------------------------------------------------------------------------------
users_with_nas_csv_data_source = CsvDataSource(
    name='users_with_nas_csv',
    csv_file_path=data_path
)

# Define Entity Objects ------------------------------------------------------------------------------------------------
user = Entity(
    name='user',
    join_keys=['id']
)

# Define FeatureView Objects -------------------------------------------------------------------------------------------
users_with_nas_csv_feature_view = FeatureView(
    name='users_with_nas_csv',
    entities=[user],
    features=[
        Feature('id', Int),
        Feature('first_name', String),
        Feature('last_name', String),
        Feature('gender', String),
        Feature('timestamp', Timestamp),
        Feature('age', Float),
        Feature('is_subscribed', Bool),
        Feature('children', Int),
    ],
    source='users_with_nas_csv_data_source'
)



"""Utilities for casting snax.ValueType to pandas.dtypes"""
import json
import re
from datetime import datetime
from typing import List, Union, Any, Optional

import pandas as pd

from snax.feature import Feature
from snax.value_type import ValueType, Null, TimestampList, BoolList, FloatList, IntList, StringList, Timestamp, Bool, \
    Float, Int, String, Unknown


def safe_cast(cast_fn):
    def safe_cast_fn(value, *args, **kwargs):
        if pd.isna(value):
            return None
        try:
            return cast_fn(value, *args, **kwargs)
        except:
            return None

    return safe_cast_fn


@safe_cast
def _cast_unknown(value: Any) -> Any:
    return value


@safe_cast
def _cast_string(value: Any) -> str:
    return str(value)


@safe_cast
def _cast_int(value: Any) -> Union[float, int]:
    return int(float(value))


@safe_cast
def _cast_float(value: Any) -> float:
    return float(value)


@safe_cast
def _cast_bool(value: Any) -> bool:
    if isinstance(value, str):
        if value.lower() in ['true', 't', '1', '1.0']:
            return True
        if value.lower() in ['false', 'f', '0', '0.0']:
            return False

    return bool(value)


@safe_cast
def _cast_timestamp(value: Any, format_string: Optional[str] = None) -> datetime:
    if isinstance(value, str):
        return datetime.strptime(value, format_string)
    elif isinstance(value, pd.Timestamp):
        return value.to_pydatetime()
    else:
        raise ValueError(f'Cannot cast {value} to timestamp')


def _strlist_to_list(value: str) -> List[str]:
    return json.loads(value)


@safe_cast
def _cast_string_list(value: Any) -> List[str]:
    raw_values = _strlist_to_list(value)
    return [_cast_string(value) for value in raw_values]


@safe_cast
def _cast_int_list(value: Any) -> List[Union[float, int]]:
    raw_values = _strlist_to_list(value)
    return [_cast_int(value) for value in raw_values]


@safe_cast
def _cast_float_list(value: Any) -> List[float]:
    raw_values = _strlist_to_list(value)
    return [_cast_float(value) for value in raw_values]


@safe_cast
def _cast_bool_list(value: Any) -> List[bool]:
    raw_values = _strlist_to_list(value)
    return [_cast_bool(value) for value in raw_values]


@safe_cast
def _cast_timestamp_list(value: Any, format_string: str) -> List[datetime]:
    raw_values = _strlist_to_list(value)
    return [_cast_timestamp(value, format_string=format_string) for value in raw_values]


@safe_cast
def _cast_null(value: Any) -> Any:
    return value


_FEATURE_TYPE_TO_CAST_FN = {
    Unknown: _cast_unknown,
    String: _cast_string,
    Int: _cast_int,
    Float: _cast_float,
    Bool: _cast_bool,
    Timestamp: _cast_timestamp,
    StringList: _cast_string_list,
    IntList: _cast_int_list,
    FloatList: _cast_float_list,
    BoolList: _cast_bool_list,
    TimestampList: _cast_timestamp_list,
    Null: _cast_null,
}


def _get_first_non_missing_item(series: pd.Series) -> Any:
    for item in series:
        if not pd.isna(item):
            return item

    return None


def _is_string_type(series: pd.Series) -> bool:
    first_nonmissing_item = _get_first_non_missing_item(series)
    return isinstance(first_nonmissing_item, str)


def guess_timestamp_format(series: pd.Series) -> str:
    timestamp_format = _get_first_non_missing_item(series)

    if timestamp_format.strip().startswith('['):
        timestamp_format = _strlist_to_list(timestamp_format)[0]

    time_patterns = [
        (re.compile('[0-9]{2}:[0-9]{2}:[0-9]{2}.[0-9]+'), '%H:%M:%S.%f'),
        (re.compile('[0-9]{2}:[0-9]{2}:[0-9]{2}'), '%H:%M:%S'),
    ]

    for time_pattern_regex, time_format in time_patterns:
        timestamp_format = time_pattern_regex.sub(time_format, timestamp_format)

    date_patterns = [
        (re.compile('[0-9]{4}-[0-9]{2}-[0-9]{2}'), '%Y-%m-%d'),
        (re.compile('[0-9]{2}.[0-9]{2}.[0-9]{4}'), '%d.%m.%Y'),

    ]
    for date_pattern_regex, date_format in date_patterns:
        timestamp_format = date_pattern_regex.sub(date_format, timestamp_format)

    return timestamp_format


def cast_to_feature_type(series: pd.Series, feature_type: ValueType) -> pd.Series:
    cast_fn = _FEATURE_TYPE_TO_CAST_FN.get(feature_type, _cast_unknown)
    kwargs = dict()

    if feature_type in [TimestampList, Timestamp] and _is_string_type(series):
        kwargs['format_string'] = guess_timestamp_format(series)

    def cast_fn_kwargs(value):
        return cast_fn(value, **kwargs)

    return series.apply(cast_fn_kwargs)


def cast_to_feature_types(dataframe: pd.DataFrame, features: List[Feature]) -> pd.DataFrame:
    dataframe = dataframe.copy()
    for feature in features:
        dataframe[feature.name] = cast_to_feature_type(dataframe[feature.name], feature.dtype)

    return dataframe
